package sun.util.locale;

import java.lang.ref.SoftReference;
import java.util.concurrent.ConcurrentHashMap;
import java.lang.ref.ReferenceQueue;
import java.util.concurrent.ConcurrentMap;

public abstract class LocaleObjectCache<K, V>
{
    private ConcurrentMap<K, LocaleObjectCache.CacheEntry<K, V>> map;
    private ReferenceQueue<V> queue;
    
    public LocaleObjectCache() {
        this(16, 0.75f, 16);
    }
    
    public LocaleObjectCache(int initialCapacity, float loadFactor, int concurrencyLevel) {
        this.queue = new ReferenceQueue();
        this.map = new ConcurrentHashMap(initialCapacity, loadFactor, concurrencyLevel);
    }
    
    public V get(K k) {
        Object o = null;
        this.cleanStaleEntries();
        LocaleObjectCache.CacheEntry<Object, Object> cacheEntry = (LocaleObjectCache.CacheEntry)this.map.get(k);
        if (cacheEntry != null) {
            o = cacheEntry.get();
        }
        if (o == null) {
            Object object = this.createObject(k);
            Object normalizeKey = this.normalizeKey(k);
            if (normalizeKey == null || object == null) {
                return null;
            }
            LocaleObjectCache.CacheEntry o2 = new LocaleObjectCache.CacheEntry(normalizeKey, object, this.queue);
            LocaleObjectCache.CacheEntry<K, V> cacheEntry2 = (LocaleObjectCache.CacheEntry<K, V>)this.map.putIfAbsent(normalizeKey, o2);
            if (cacheEntry2 == null) {
                o = object;
            }
            else {
                o = cacheEntry2.get();
                if (o == null) {
                    this.map.put(normalizeKey, o2);
                    o = object;
                }
            }
        }
        return o;
    }
    
    protected V put(K k, V v) {
        LocaleObjectCache.CacheEntry<Object, Object> cacheEntry = (LocaleObjectCache.CacheEntry)this.map.put(k, new LocaleObjectCache.CacheEntry(k, v, this.queue));
        return (cacheEntry == null) ? null : cacheEntry.get();
    }
    
    private void cleanStaleEntries() {
        LocaleObjectCache.CacheEntry<Object, Object> cacheEntry;
        while ((cacheEntry = (LocaleObjectCache.CacheEntry)this.queue.poll()) != null) {
            this.map.remove(cacheEntry.getKey());
        }
    }
    
    protected abstract V createObject(K p0);
    
    protected K normalizeKey(K k) {
        return k;
    }
    
    private static class CacheEntry<K, V> extends SoftReference<V>
    {
        private K key;
        
        CacheEntry(K key, V referent, ReferenceQueue<V> q) {
            super(referent, q);
            this.key = key;
        }
        
        K getKey() {
            return this.key;
        }
    }
}
